/*		Name:		chump.c
		Version:	0.0.12
		Date:		29/8/2002
		(Assembly and) Dissassembly functions 
*/

/* This code is property of Charles Brej  and is used in this project 
									with his permission */
#include "chump.h"
#include "dotparse.h"
#include <glib.h>
#include <string.h>
#include <ctype.h>

const AsmStatus asm_default_status = {NULL, 0, 0};  // Symbol* symbol_table, int symbol_count, address;



DefinitionStack *asm_parser (GList * list)          // Pass the parsed list to tis function with the name as the first element
{                                                   // isa     "ARM32"  (define "reg" ....
	if (!SCAN_IS_STRING (SCAN_LIST_FIRST (list)))   //      ^ start
	{
        g_print ("line:%d position:%d  Expected string to define the disassembly\n",
            SCAN_LIST_FIRST (list)->line, SCAN_LIST_FIRST (list)->position);
        return NULL;
	}
	return asm_define (list->next, NULL, SCAN_STRING (SCAN_NODE (list->data)));
}


DefinitionStack *asm_define (GList * parselist, DefinitionStack * deflist, char *name)
{               //          the list of defines,  a stack of prevous defines,  name of current define
    DefinitionStack *definitionelement = g_new (DefinitionStack, 1);  // create our new define
    DefinitionStack *definitionstack = definitionelement;             // this is the stack with our new element on top
    DefinitionStack *deflisttemp;
    GList *sublist;

    definitionelement->next = deflist;
    definitionelement->rules = NULL;
    definitionelement->string = name;
    
    while (parselist)                                               // walk through the input list
	{


		if (SCAN_IS_LIST ((PtrSCANNode) parselist->data))           // form (define "X" (...
		{
			sublist = ((PtrSCANNode) parselist->data)->body.list;   // take a look at the sublist
			if (!sublist)
			{
				g_print ("line:%d position:%d  Empty list when defining '%s'\n",    // sublist is empty!!! panic!!!
					((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position, name);
				return definitionelement;
			}

			if (!sublist->next)                                          // sublist has only one element!!! panic!!!
			{
				g_print ("line:%d position:%d  List of only 1 element when defining '%s'\n",
					((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position, name);
				return definitionelement;
			}

			if (SCAN_IS_SYMBOL ((PtrSCANNode) sublist->data) &&     // form (define "X" (define ...
				SCAN_IS_STRING ((PtrSCANNode) sublist->next->data) &&
				!g_strcasecmp (((PtrSCANNode) sublist->data)->body.string, DEFINITIONSTRING))
				        definitionstack = asm_define (sublist->next->next, definitionstack,     // stick the define on the stack 
					        ((PtrSCANNode) sublist->next->data)->body.string);                  // recursively
			else if (SCAN_IS_SYMBOL ((PtrSCANNode) sublist->data) &&     // form (define "X" (enum ...
				      SCAN_IS_NUMBER ((PtrSCANNode) sublist->next->data) &&
				      !g_strcasecmp (((PtrSCANNode) sublist->data)->body.string, ENUMERATESTRING)) {
                int size, number, temp;
                Rule *rule;
                RuleElement *element;
                number=0;
                sublist = sublist->next;
                size = SCAN_NUMBER(SCAN_LIST_FIRST(sublist));
                while ((sublist = sublist->next)){
                    if (!SCAN_IS_STRING(SCAN_LIST_FIRST(sublist))){
                        g_print ("line:%d position:%d  Enum lists should be created using only \"strings\" \n",
					                    ((PtrSCANNode) sublist->data)->line, ((PtrSCANNode) sublist->data)->position);
                        break;
                        }
                    rule = g_new(Rule,1);
                    rule->type = TRANRULE;
                    element = g_new(RuleElement,1);
                    element->string = SCAN_STRING(SCAN_LIST_FIRST(sublist));
                    element->type = TERM;
                    element->twin=0;
                    element->twinsplit=0;
                    element->split=NULL;
                    element->rules=NULL;
                    rule->asci = g_list_append(NULL,element);
                    element = g_new(RuleElement,1);
                    element->string = g_new(char, size+1);
                    temp=size;
                    element->string[size]='\0';
                    while (temp--){
                        if (number & 1<<temp)   element->string[size-temp-1]='I';
                        else                    element->string[size-temp-1]='O';
                        }
                    element->type = TERM;
                    element->twin=0;
                    element->twinsplit=0;
                    element->split=NULL;
                    element->rules=NULL;
                    rule->bit = g_list_append(NULL,element);
    				definitionelement->rules = g_list_append (definitionelement->rules, rule);      // add it to the rules
                    number++;
                    }
                }
                        
            else                                                    // form (define "X" (( Y ...
				definitionelement->rules = g_list_append (definitionelement->rules,       // add it to the rules
                                                        asm_rule (sublist, definitionstack));
		}


		if (SCAN_IS_SYMBOL ((PtrSCANNode) parselist->data))         // form (define "X" Y)
		{
			deflisttemp = definitionstack;
			while (g_strcasecmp (((PtrSCANNode) parselist->data)->body.string, deflisttemp->string))    // fook for the name of the stack
			{
				deflisttemp = deflisttemp->next;
				if (!deflisttemp)                       // we go to the end of the list and still didnt find it! panic!!!
				{
					g_print ("line:%d position:%d  '%s' has not been defined\n",
						((PtrSCANNode) parselist->data)->line,
						((PtrSCANNode) parselist->data)->position, ((PtrSCANNode) parselist->data)->body.string);
					return definitionelement;
				}
			}
			definitionelement->rules = g_list_concat (definitionelement->rules, g_list_copy (deflisttemp->rules));
		}                                           // copy the rules from the original to the current rules
		parselist = parselist->next;                // NEXT!
	}

	                    //should free all elements above the definition stack
	deflisttemp = definitionstack;
	while (definitionstack != definitionelement)
	{
		deflisttemp = definitionstack;
		definitionstack = definitionstack->next;        // do it proper!! losing a bit or mem during definition time. fixme
		g_free (deflisttemp);
	}

	return definitionelement;
}

Rule *asm_rule (GList * parselist, DefinitionStack * deflist)
//              parsed list      , definition stack 
{   //      form ((ascii)(binary))  or  ( int 7 + 4 )

	Rule *rule = g_new (Rule, 1);   //make new rule
	RuleElement *belement, *telement;
	GList *text;
	GList *binary;
	DefinitionStack *deflisttemp;
	GList *binarytemp;
	char *legaltest;
    SplitElement *split;

	rule->asci = NULL;      // initialise new rule
	rule->bit = NULL;
	rule->type = TRANRULE;
    
	if (SCAN_IS_SYMBOL ((PtrSCANNode) parselist->data))     // form ( int ...
	{
		if (!g_strcasecmp (((PtrSCANNode) parselist->data)->body.string, INTIGERSTRING))    // test for int, relative ...
			rule->type = INTRULE;
		if (!g_strcasecmp (((PtrSCANNode) parselist->data)->body.string, RELATIVESTRING))
			rule->type = RELATIVERULE;
		if (!g_strcasecmp (((PtrSCANNode) parselist->data)->body.string, UINTIGERSTRING))
			rule->type = UINTRULE;
		if (!g_strcasecmp (((PtrSCANNode) parselist->data)->body.string, URELATIVESTRING))
			rule->type = URELATIVERULE;
            
		if (TRANRULE == rule->type)             // didnt match any of them!!! panic!!!
		{
			g_print ("line:%d position:%d  Only 'int' and 'relative' are valid rules \n",
				((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position);
			return rule;
		}
		parselist = parselist->next;           // start working through operands

		if (!SCAN_IS_NUMBER ((PtrSCANNode) parselist->data)) // no size specified!!! panic!!!
		{
			g_print ("line:%d position:%d  Int and float rules need a size as the first operand \n",
				((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position);
			return rule;
		}

		rule->bit = g_list_append (rule->bit, GINT_TO_POINTER (((PtrSCANNode) parselist->data)->body.number));
		parselist = parselist->next;            // stick size as the first element of bitlist
		while (parselist)                       // walk through rest
		{
			if (!parselist->next)               // if each operand needs a number!!! panic!!!
			{
				g_print
					("line:%d position:%d  Int and float rules need mathematical operations in sepperated pairs eg: + 4 * 8\n",
					((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position);
				return rule;
			}

			if (!SCAN_IS_SYMBOL ((PtrSCANNode) parselist->data) ||
				!SCAN_IS_NUMBER ((PtrSCANNode) parselist->next->data))      // ensure its an operator followed by a number
			{
				g_print                                                     // else panic!!!
					("line:%d position:%d  Int and float rules need mathematical operations in sepperated pairs eg: + 4 * 8\n",
					((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position);
				return rule;
			}

			rule->asci = g_list_append (rule->asci, ((PtrSCANNode) parselist->data)->body.string);  // stick the symbol and number on the lists
			rule->bit = g_list_append (rule->bit, GINT_TO_POINTER (((PtrSCANNode) parselist->next->data)->body.number));
			parselist = parselist->next->next;      // move two forward
		}
		return rule;
	}

                                // form ((asm)(bin))
	rule->type = TRANRULE;
	if (parselist->next->next ||
		!SCAN_IS_LIST ((PtrSCANNode) parselist->data) || !SCAN_IS_LIST ((PtrSCANNode) parselist->next->data))
	{
		g_print ("line:%d position:%d  A rule must have two LISTS! \n",
			((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position);
		return rule;
	}                       // ensure two lists exist else panic!!!
    
	text = SCAN_NODE (parselist->data)->body.list;
	binary = SCAN_NODE (parselist->next->data)->body.list;
	while (binary)          // walk throug binary list
	{
	    belement = g_new (RuleElement, 1);      //create an element
	    if (!SCAN_IS_SYMBOL (SCAN_NODE (binary->data)))     // binary sections have to be made from symbols
		{                                                   // if the're not then panic!!!
			g_print
				("line:%d position:%d  A rule binary field must be made from characters IOXZ or a number starting with '#'\n",
				((PtrSCANNode) binary->data)->line, ((PtrSCANNode) binary->data)->position);
			return rule;
		}
		belement->type = TERM;                                      // make it into a term for now
		belement->string = SCAN_NODE (binary->data)->body.string;
		rule->bit = g_list_append (rule->bit, (gpointer) belement); // add it to the list
		binary = binary->next;
	}

	while (text)                    // now walk through the text
	{
		telement = g_new (RuleElement, 1);          // add a new element
		rule->asci = g_list_append (rule->asci, (gpointer) telement);       // add it to the list

		switch (((PtrSCANNode) text->data)->type)       // what type of element is it?
		{
        
		case SCANString:                                // its a terminal eg "add"
			telement->string = SCAN_NODE (text->data)->body.string;
			telement->type = TERM;                      // just add the string then 
			break;

		case SCANList:                                  // its an inline definition
            telement->type = SPLITNONTERM;
            telement->split = NULL;
			if (!SCAN_LIST(SCAN_LIST_FIRST(text)))
			{
				g_print ("line:%d position:%d  Rule must have a body \n",        // ensure the list has a body, else panic!!!
					((PtrSCANNode) text->data)->line, ((PtrSCANNode) text->data)->position);
				return rule;
			}
			if (SCAN_IS_SYMBOL (SCAN_LIST_FIRST(SCAN_LIST(SCAN_LIST_FIRST(text)))))   // ensure the definition is in quotes, else panic!!!
			{
				g_print ("line:%d position:%d  '%s' should be a in quotes\n",
					((PtrSCANNode) text->data)->line,
					((PtrSCANNode) text->data)->position,
					(SCAN_NODE (SCAN_NODE (text->data)->body.list->data))->body.string);
				return rule;
			}
            
 			if (SCAN_IS_NUMBER (SCAN_LIST_FIRST(SCAN_LIST(SCAN_LIST_FIRST(text)))))   // ensure the definition is in quotes, else panic!!!
			{
				g_print ("line:%d position:%d  A rule sould start with a string and not a number\n",
					((PtrSCANNode) text->data)->line,
					((PtrSCANNode) text->data)->position);
				return rule;
			}
           
            

			if (SCAN_IS_LIST (SCAN_LIST_FIRST(SCAN_LIST(SCAN_LIST_FIRST(text)))))  
			{
                binary = SCAN_LIST(SCAN_LIST_FIRST(SCAN_LIST(SCAN_LIST_FIRST(text))));
                if (!binary) {
        				g_print ("line:%d position:%d  The rule entry seems to be an empty list! I have no idea what to do with this!\n",
        					SCAN_LIST_FIRST(text)->line,
		        			SCAN_LIST_FIRST(text)->position);
				        return rule;
            			}
                
                telement->string = "";              // define the string as null as its a split
    			deflisttemp = asm_define (SCAN_LIST_REST(SCAN_LIST(SCAN_LIST_FIRST(text))), deflist, "");  // work out its defenition 
			    telement->rules = deflisttemp->rules;
                g_free(deflisttemp);


                while (binary) {                // walk through the split list
                        if (!SCAN_IS_LIST (SCAN_LIST_FIRST(binary))) {
            				g_print ("line:%d position:%d  The rule split sub list seems to have non list entrys\n",
            					((PtrSCANNode) binary->data)->line,
		            			((PtrSCANNode) binary->data)->position);
				            return rule;
            			    }
                            
                        if (!SCAN_IS_STRING(SCAN_LIST_FIRST(SCAN_LIST(SCAN_LIST_FIRST(binary))))) {
            				g_print ("line:%d position:%d  The split rule parts should be in form: name string followed by size number\n",
            					((PtrSCANNode) binary->data)->line,
		            			((PtrSCANNode) binary->data)->position);
				            return rule;
            			    }
                            
                        if (!SCAN_IS_NUMBER(SCAN_LIST_SECOND(SCAN_LIST(SCAN_LIST_FIRST(binary))))) {
            				g_print ("line:%d position:%d  The split rule parts should be in form: name string followed by size number\n",
            					((PtrSCANNode) binary->data)->line,
		            			((PtrSCANNode) binary->data)->position);
				            return rule;
            			    }
                            
                        
                        split = g_new(SplitElement,1);
                        telement->split= g_list_append (telement->split, split);

                        split->string = SCAN_STRING(SCAN_LIST_FIRST (SCAN_LIST(SCAN_LIST_FIRST(binary))));
                        split->size =   SCAN_NUMBER(SCAN_LIST_SECOND(SCAN_LIST(SCAN_LIST_FIRST(binary))));
                        split->type =   SPLIT;

        	    		binarytemp = rule->bit;
        		    	while (g_strcasecmp (split->string, ((RuleElement *) binarytemp->data)->string)) {   // find its twin in the binary section
		        		    
                            if (!binarytemp->next)  {                               // we havent found it! panic!!!
	                            g_print ("line:%d position:%d  '%s' needs to be in the binary section too\n",
		    		                ((PtrSCANNode) text->data)->line, ((PtrSCANNode) text->data)->position, telement->string);
			    	            return rule;
        				        }
                            
            				binarytemp = binarytemp->next;          //next!
	            		    }

                       
  		    	        split->twin = g_list_position (rule->bit, binarytemp);

                        ((RuleElement *) binarytemp->data)->twin = g_list_index (rule->asci, telement);     //connect the twins 
                        ((RuleElement *) binarytemp->data)->twinsplit = g_list_index (telement->split, split);
                        ((RuleElement *) binarytemp->data)->type = NONTERM;
                        ((RuleElement *) binarytemp->data)->rules = telement->rules;
                        
                        binary=binary->next;
                        }
                break;
			}


			if (SCAN_IS_STRING (SCAN_LIST_FIRST(SCAN_LIST(SCAN_LIST_FIRST(text))))) {
                telement->string = SCAN_NODE (SCAN_NODE (text->data)->body.list->data)->body.string;            // define the string 
    			deflisttemp = asm_define (SCAN_LIST_REST(SCAN_LIST(SCAN_LIST_FIRST(text))), deflist, telement->string);  // work out its defenition 

	    		binarytemp = rule->bit;
		    	while (g_strcasecmp (telement->string, ((RuleElement *) binarytemp->data)->string)) {   // find its twin in the binary section
				    if (!binarytemp->next) {                                // we havent found it! panic!!!
	                    g_print ("line:%d position:%d  '%s' needs to be in the binary section too\n",
		    		        ((PtrSCANNode) text->data)->line, ((PtrSCANNode) text->data)->position, telement->string);
			    	    return rule;
    				    }
    				binarytemp = binarytemp->next;          //next!
    	    		}
                    
		    	((RuleElement *) binarytemp->data)->type = NONTERM;     // correct the binary definition as its a 'nonterm'
			    telement->type = NONTERM;                               // as is our current element
    			telement->split = NULL;
	    		((RuleElement *) binarytemp->data)->split = NULL;
		    	((RuleElement *) binarytemp->data)->rules = deflisttemp->rules; //deflisttemp->rules are our rules for this nonterm
			    telement->rules = deflisttemp->rules;                           // on both sides
    			((RuleElement *) binarytemp->data)->twin = g_list_index (rule->asci, telement);     //connect the twins
                ((RuleElement *) binarytemp->data)->twinsplit = 0;
		    	telement->twin = g_list_position (rule->bit, binarytemp);                           // together
			    g_free (deflisttemp);                                       // this desnt really cut it. fixme
    			break;
                }

		case SCANSymbol:                                                // its a simple nonterm 
			telement->string = SCAN_NODE (text->data)->body.string;
			deflisttemp = deflist;
			while (g_strcasecmp (telement->string, deflisttemp->string))    // find its definition on the stack
			{
				deflisttemp = deflisttemp->next;
				if (!deflisttemp)
				{
					g_print ("line:%d position:%d  '%s' has not been defined\n",        // no definition found. panic!!!
						((PtrSCANNode) text->data)->line, ((PtrSCANNode) text->data)->position, telement->string);
					return rule;
				}
			}

			binarytemp = rule->bit;
			while (g_strcasecmp (telement->string, ((RuleElement *) binarytemp->data)->string))
			{
				binarytemp = binarytemp->next;                          // find its twin
				if (!binarytemp)
				{
					g_print ("line:%d position:%d  '%s' needs to be in the binary section too\n",   // no twin found
						((PtrSCANNode) text->data)->line, ((PtrSCANNode) text->data)->position, telement->string);
					return rule;
				}
			}
			((RuleElement *) binarytemp->data)->type = NONTERM;     // correct the binary definition as its a 'nonterm'
			telement->type = NONTERM;                               // as is our current element
			((RuleElement *) binarytemp->data)->rules = deflisttemp->rules; //deflisttemp->rules are our rules for this nonterm
			telement->rules = deflisttemp->rules;                           // on both sides
			((RuleElement *) binarytemp->data)->twin = g_list_index (rule->asci, telement);     //connect the twins
			telement->twin = g_list_position (rule->bit, binarytemp);                           // together
            
			break;
		default:
			g_print ("line:%d position:%d  unexpected thing!\n",            // well its not a list string or symbol so panic!!!
				((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position);
			return rule;
		}
		text = text->next;      // next text!
	}
    
    
	binary = rule->bit;
	while (binary)              // walk through the binary (again sortof)
	{
		if (TERM == ((RuleElement *) binary->data)->type)       //if its a term then ensure its a good one
		{
			legaltest = ((RuleElement *) binary->data)->string; 
			if ('#' == *legaltest)
			{
				legaltest++;
				((RuleElement *) binary->data)->string = legaltest; // if its starts with a # then advance by 1
			}
			while (*legaltest)          // convert 1s to Is and 0s to Os
			{
				if ('1' == *legaltest)
					*legaltest = 'I';
				if ('0' == *legaltest)
					*legaltest = 'O';
				legaltest++;
			}
			legaltest = ((RuleElement *) binary->data)->string;     // test if all characters are legal
			while (*legaltest)
			{
				if ('I' != *legaltest && 'O' != *legaltest && 'Z' != *legaltest && 'X' != *legaltest)  // if not valid then panic!!!
				{
					g_print
						("line:%d position:%d  This rule's binary field contains '%s' which is not a valid binary value nor is it in the text field\n",
						((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position,
						((RuleElement *) binary->data)->string);
					break;
				}
				legaltest++;
			}
		}
		binary = binary->next; // next!
	}
	return rule;        //when donr return rule
}




Disasm asm_disassemble_number (GList * bitfield, GList * rules, AsmStatus stat)
//                  list of binary, list of rules, address (should be bigger. fixme)
{
 Disasm ret;
 Disasm error;
 GList *binary;
 char *c;
 RuleElement *newelement;
 unsigned int numbera;
 unsigned int size;
    
 error.distext = &dis_error;         // error dissasembly
 error.bitfield = &dis_error;
 ret.bitfield = bitfield;
 




 if ((int) ((Rule *) rules->data)->bit->data > g_list_length (bitfield)) return error;  // if its longer than the numbera
 binary = g_list_nth (ret.bitfield, (int) ((Rule *) rules->data)->bit->data - 1);        // of bits fetched then error out
 numbera = 0;                                                   // walk through the binary backwards
 
 if (GPOINTER_TO_INT (binary->data)){                        // if its signed then the first bit means the sign
    if (    UINTRULE      == ((Rule *) rules->data)->type
         || URELATIVERULE == ((Rule *) rules->data)->type)  numbera=1;
    else                                                    numbera=-1; // numbera is the accumulator
    }
    
 while (binary != ret.bitfield) {   // walk backwards
    binary = binary->prev;
    numbera = numbera << 1;
    if (GPOINTER_TO_INT (binary->data)) numbera++;   // keep adding and shifting
	}

 {
    GList *number = g_list_last (((Rule *) rules->data)->bit);      // apply operations from end 
    GList *opp    = g_list_last (((Rule *) rules->data)->asci);
    int imm;
    while(opp){                                 // all opertions
        imm=GPOINTER_TO_INT (number->data);     // imm is the operand
        c= (char*)opp->data;                    // c points to the operator
        switch (*c){							// if not inverted then just do normal inverse operations
            case '+':
                numbera -=imm;
                if (UINTRULE      == ((Rule *) rules->data)->type)
                    if ((int)numbera<0) return error;
                break;
                    
            case '~':
                numbera = imm - numbera;
                break;

            case '-':
                numbera +=imm;
                break;
            
            case '*':
                numbera /=imm;
                break;
                    
            case '/':
                numbera *=imm;
                break;
            
            case '>':
                numbera = numbera << imm;
                break;

            case '<':
                numbera = numbera >> imm;
                break;
                                
            case '@':
                c++;
                size = g_strtod(c,NULL);
                numbera = (((numbera >> imm) | (numbera << (size - imm))) & ((((1 << (size-1) )-1)<<1)+1 ));
                break;

            default:
                break;
            }
							
        opp=opp->prev;              // next operation!
        number=number->prev;
        }
 }
 if (    (RELATIVERULE == ((Rule *) rules->data)->type)      // if its a relative rule then add the address
     || (URELATIVERULE == ((Rule *) rules->data)->type)) numbera += stat.address;

 ret.bitfield = g_list_nth (ret.bitfield, (int) ((Rule *) rules->data)->bit->data); // advance bitfield forward
 
 
 newelement = g_new (RuleElement, 1); 

 c = asm_symboltable_find_address (stat.symbol_table, stat.symbol_count, numbera);
 if (c && (    (RELATIVERULE == ((Rule *) rules->data)->type)     
           || (URELATIVERULE == ((Rule *) rules->data)->type))) 
        newelement->string = g_strdup(c);
 else   newelement->string = g_strdup_printf ("%X", numbera);    // make a new element which will be freed later
 
 newelement->type = TERM2FREE;                           // and sick our number into it
 ret.distext = g_list_append (NULL, newelement);         // add it to the return text text
 return ret;                                             // and return
}



Disasm asm_disassemble (GList* bitfield, GList* rules, AsmStatus stat)
//                  list of binary, list of rules, address (should be bigger. fixme)
{
	Disasm ret;
	Disasm rep;
	Disasm error;
	GList *binary;
	char *c;
	RuleElement *newelement;
	unsigned int flag;
    unsigned int size;
    
	error.distext = &dis_error;         // error dissasembly
	error.bitfield = &dis_error;

	while (rules) {                     // walk through rules given
      if (TRANRULE != ((Rule *) rules->data)->type) {  // its not a number rule
            rep = asm_disassemble_number (bitfield, rules, stat);
  	        if (&dis_error != rep.bitfield)      // if replay was OK then return
       	        return rep;
            }
      else {
		ret.bitfield = bitfield;
		binary = g_list_last (((Rule *) rules->data)->bit);         // these are translation rules
		ret.distext = g_list_copy (((Rule *) rules->data)->asci);   // ret.distext is now a copy of the contents of the translation

		while (1){
			flag = 0;
			if (!binary) return ret;                                    // if no binary then we are done
			if (NONTERM == ((RuleElement *) binary->data)->type) {      // its a non-terminal
                if (SPLITNONTERM == ((RuleElement *)g_list_nth_data (ret.distext, ((RuleElement *) binary->data)->twin))->type){
					newelement = g_new (RuleElement, 1);            // make a new rule
					newelement->rules = NULL; 
//                    g_print("%x %x %x\n",newelement->rules->data, newelement->rules->next, newelement->rules->prev);
					newelement->type = SPLITNONTERM2FREE;                // stick it into the reply
                    newelement->split = g_list_copy(((RuleElement *)g_list_nth_data (ret.distext, ((RuleElement *) binary->data)->twin))->split);
					(g_list_nth (ret.distext, ((RuleElement *) binary->data)->twin))->data = (gpointer) newelement;
//                    g_print("1\n");
                    }

                if (SPLITNONTERM2FREE == ((RuleElement *)(g_list_nth_data (ret.distext, ((RuleElement *) binary->data)->twin)))->type){
                    SplitElement* newsplit = g_new(SplitElement,1);
                    RuleElement* rule = (RuleElement*) g_list_nth_data (ret.distext, ((RuleElement *) binary->data)->twin);
                    GList* splitlist, *splitlist2;
                    newsplit->type = SPLIT2FREE;
                    newsplit->size = ((SplitElement*)  g_list_nth_data (rule->split, ((RuleElement *) binary->data)->twinsplit))->size;
                    newsplit->string = ((SplitElement*)  g_list_nth_data (rule->split, ((RuleElement *) binary->data)->twinsplit))->string;
                    newsplit->bin = NULL;
                    g_list_nth (rule->split, ((RuleElement *) binary->data)->twinsplit)->data = newsplit;
                    size = newsplit->size;
                    while(size--){
                        if (!ret.bitfield) {flag=1; g_print("a\n");break;}
                        newsplit->bin = g_list_append(newsplit->bin, ret.bitfield->data);
        	            ret.bitfield = ret.bitfield->next;
                        }
                    if (flag) break;
                    splitlist=rule->split;
                    while (SPLIT2FREE == ((SplitElement*)splitlist->data)->type) {
                        splitlist=splitlist->next;
                        if (!splitlist) {break;}
                        }
                    if (!splitlist) {
                        splitlist=rule->split;
                        splitlist2=NULL;
                        while (splitlist) {
                            splitlist2 = g_list_concat(((SplitElement*)splitlist->data)->bin, splitlist2);
                            ((SplitElement*)splitlist->data)->bin=NULL;
                            splitlist=splitlist->next;
                            }
                        rep = asm_disassemble (splitlist2, ((RuleElement *) binary->data)->rules, stat);   // in that case disassable it
                        g_list_free(splitlist2);
    	    		    if (&dis_error != rep.bitfield && !rep.bitfield)    // if replay was OK then
		        	        rule->rules = rep.distext;                      // make a nonterm which will be freed
                        
                        else { flag = 1;}                                    // if replay was an error then flag for try next rule
                        }
                    }
                else {
                 
    				rep = asm_disassemble (ret.bitfield, ((RuleElement *) binary->data)->rules, stat);   // in that case disassable it
	    			if (&dis_error != rep.bitfield) {                   // if replay was OK then
    					ret.bitfield = rep.bitfield;                    // advance bitfield
	    				newelement = g_new (RuleElement, 1);            // make a new rule
		    			newelement->rules = rep.distext;                // make a nonterm which will be freed
			    		newelement->type = NONTERM2FREE;                // stick it into the reply
				    	(g_list_nth (ret.distext, ((RuleElement *) binary->data)->twin))->data = (gpointer) newelement;
        				}
                    else  flag = 1;                                       // if replay was an error then flag for try next rule
                    }
			}
            else {
				c = (char *) ((RuleElement *) binary->data)->string;    // if it was a terminal
				while (*c)                                              // goto end of the binary term
					c++;
                if (ret.bitfield)                                       // 
				 while (((RuleElement *) binary->data)->string != c)    // ensure it matches
				 {
					c--;
					if (ret.bitfield){
                     if ('I' == *c && !GPOINTER_TO_INT (ret.bitfield->data))    // if doesnt match then flag=1
						 flag = 1;
					 if ('O' == *c &&  GPOINTER_TO_INT (ret.bitfield->data))
						 flag = 1;
					 ret.bitfield = ret.bitfield->next;
                     }
                    else flag = 1;          // if ret.bitfield runs out then try another
				 }
                else flag = 1;
                 
		        }
		    if (flag) break;                    // if flag is up then break free and try next rule
		    binary = binary->prev;              // else keep going through current
            }
        asm_freestringlist (ret.distext);
        }
	rules = rules->next;
	}
  return error;                       // if no rules matched then error
}




Asm* asm_assemble_number (char* text, GList* rules, AsmStatus stat, int last)
//               input text, rule list, addess, is ths the last rule?
{
 Asm *ret = g_new(Asm,1);
 RuleElement *newelement;
 int flag;
 unsigned int size;
 unsigned int value;
 int index;
 char* start;
 char* c;

 ret->text = text;

 while (' '  == *ret->text) ret->text++;         // skip spaces
 flag=0;
 
 size = asm_getsymbol_length(ret->text);
 if (size) {
  c = g_strndup(ret->text, size);
  index = asm_findstrinsymboltable(c, stat.symbol_table, stat.symbol_count);
  g_free (c);
  }
 else index = -1;

 g_print("symbol no:%d\n",  value);
 
 if (0 > index){
    value=0;
    if (!isxdigit(*ret->text)) {g_free(ret); return NULL;}            // if its not a number then don't bother
    while (isxdigit(*ret->text)){                   // walk through the number
        if (*ret->text< 0x3A)    value = (value<<4) + *ret->text - 0x30;
        else                     value = (value<<4) + (*ret->text & 7) + 9;
        ret->text++;
        g_print("hex:%x\n", value);
        }
    }
 else {
    value = stat.symbol_table[index].value;
    ret->text += size;
    g_print("symbol:%x\n", index);
    }

 if (    (RELATIVERULE == ((Rule *) rules->data)->type)
     || (URELATIVERULE == ((Rule *) rules->data)->type)) value -= stat.address; // if relative subtract address

 {
    GList *number = ((Rule *) rules->data)->bit->next;  // walk through the operations
    GList *opp = ((Rule *) rules->data)->asci;
    int imm;
    while(opp){
        imm=GPOINTER_TO_INT (number->data);
        c= (char*)opp->data;
        
        switch (*c){
            case '~':
                value = imm - value;
                break;
                    
            case '+':
                value +=imm;
                break;
    
            case '-':
                value -=imm;
                break;
            
            case '*':
                value *=imm;
                break;
    
            case '/':
                value /=imm;
                break;
            
            case '>':
                value = value >> imm;
                break;
                
            case '<':
                value = value << imm;
                break;
                
            case '@':
                c++;
                size = g_strtod(c,NULL);
                value = ((value << imm) | ((value >> (size - imm)) & ((((1 << (size-1) )-1)<<1)+1 )));
                break;
                
            default:
                break;
            }
        opp=opp->next;
        number=number->next;
        }
 }
 
 if ((value < (1<<(unsigned int) ((Rule *) rules->data)->bit->data)) ||
                   (RELATIVERULE == ((Rule *) rules->data)->type) ||
                        (INTRULE == ((Rule *) rules->data)->type))
    if (!flag) {											            // flag is set if something went wrong
        start = g_new(char,(int) ((Rule *) rules->data)->bit->data + 1);//make a new string
        c = start + (int) ((Rule *) rules->data)->bit->data;			//terminate it with a /0
        *c = '\0';
        while(start != c--){											// write the string
            if (value & 1)  *c = 'I';
            else        *c = 'O';
            value = value >> 1;
            }
        newelement = g_new (RuleElement, 1);							// make a new rule
        newelement->string = start;
        newelement->type = TERM2FREE;									// make sure its freed afterwards
        ret->binary = g_list_append (NULL, newelement); 				// return the result in a list in ret->binary
        if (last) while (' '  == *ret->text) ret->text++;				// if its last then walk through the rest of the spaces
        if (!(last && *(ret->text))) {									// unless its last and there is still chars to process then
            return ret; 										        // return it
            }
        }
  g_free(ret);        
  return NULL;
}




GList* asm_assemble (char* text, GList* rules, AsmStatus stat, int last)
//               input text, rule list, addess, is ths the last rule?
{
 Asm *ret = g_new(Asm,1);
 Asm *todo;
 GList *todolist=NULL;
 GList *retlist=NULL;
 GList *rep;
 RuleElement *newelement;
 RuleElement *newelement2;
 int flag;
 char* start;
 char* c;
    
 while (rules){             // try out all rules
        flag=1;
	    ret->text = text;
	    if (TRANRULE != ((Rule *) rules->data)->type){      // its a number rule
            Asm* numberasm= asm_assemble_number (text, rules, stat, last);
            if (numberasm) retlist = g_list_append(retlist,numberasm);						// add it to the list
            }


   
        else {															// its a translation rule
          ret->text = text;
          ret->ruletext=((Rule *) rules->data)->asci;					// keep a bit of state as to which rule we are looking at
          ret->binary = g_list_copy (((Rule *) rules->data)->bit);		// binary section is just a copy of the rule's binarys
        
		  while (1){
            flag = 0;
            start = ret->text;
            while (' '  == *(start)) start++;       					// move past the white spaces
            if (!ret->ruletext) {										// if we run out of rules
                      if (last && *(start)) 						    // if its last and there is still text 
                            {asm_freestringlist (ret->binary);break;}	// then try another
                      else {retlist = g_list_append(retlist,ret); ret = g_new(Asm,1);break;} // else stick it on the retlist and try another
                      }
            
            
		    if (NONTERM == ((RuleElement *) ret->ruletext->data)->type){	// its a nonterm
//                while (' '  == *(ret->text)) ret->text++;					// skip spaces 
				rep = asm_assemble(ret->text, ((RuleElement *) ret->ruletext->data)->rules, stat, last && !ret->ruletext->next);
				// call recursively. text, rules , address, is it last?
				
 				if (rep){
                    ret->text = ((Asm*)rep->data)->text;	// did we get a reply?
 					newelement = g_new (RuleElement, 1);	// make a new element
 					newelement->rules = ((Asm*)rep->data)->binary; // point to the binary part of the reply
 					newelement->type = NONTERM2FREE;		// ensure its freed
                    rep = rep->next;
                    while (rep) {							// is there more than one reply?
     					newelement2 = g_new (RuleElement, 1);	// make another element 
 	    				newelement2->rules = ((Asm*)rep->data)->binary;
 		    			newelement2->type = NONTERM2FREE;
                        todo=rep->data; 						// todo element
                        todo->ruletext=ret->ruletext;			// now points to the one from the one from the reply
						todo->binary=asm_copystringlist(ret->binary); // binary part of the reply is a copy of the current progress
 					    (g_list_nth (todo->binary, ((RuleElement *) todo->ruletext->data)->twin))->data = (gpointer) newelement2;
										// replace the twin of what we found with the nonterm
                        todolist=g_list_prepend(todolist,todo); // add it to the todo list
                        rep=rep->next;					// next!
                        }
 					(g_list_nth (ret->binary, ((RuleElement *) ret->ruletext->data)->twin))->data = (gpointer) newelement;
									// replace the twin with the our nonterm
				    }
                else  flag = 1; // if no reply then flag for quit
			    }

		    else if (SPLITNONTERM == ((RuleElement *) ret->ruletext->data)->type){	// its a splitnonterm
				rep = asm_assemble(ret->text, ((RuleElement *) ret->ruletext->data)->rules, stat, last && !ret->ruletext->next);
				// call recursively. text, rules , address, is it last?

 				if (rep){
                    GList* split;
                    GList* rep_first = rep;
                    ret->text = ((Asm*)rep->data)->text;
                    rep = rep->next;
                    while (rep) {							// is there more than one reply?
                        split = ((RuleElement *) ret->ruletext->data)->split;
                        start = c = asm_sprint (((Asm*)rep->data)->binary, g_strdup(""));
                        todo=rep->data; 						// todo element
                        todo->ruletext=ret->ruletext;			// now points to the one from the one from the reply
						todo->binary=asm_copystringlist(ret->binary); // binary part of the reply is a copy of the current progress
                        todolist=g_list_prepend(todolist,todo); // add it to the todo list
                        while (split){
                            int size = ((SplitElement *)split->data)->size;
                            newelement2 = g_new (RuleElement, 1);
 	    	    			newelement2->type = TERM2FREE;
                            newelement2->string = g_new (char, size + 1);
                            newelement2->string[size] = '\0';
                            while (size--)  newelement2->string[size] = c[size];
                            c += size;
     					    (g_list_nth (todo->binary, ((SplitElement *)split->data)->twin))->data = (gpointer) newelement2;
                            split=split->next;
                            }
                        rep=rep->next;					// next!
                        }
                    split = ((RuleElement *) ret->ruletext->data)->split;
                    start = c = asm_sprint (((Asm*)rep_first->data)->binary, g_strdup(""));
                    while (split){
                            int size = ((SplitElement *)split->data)->size;
                            newelement2 = g_new (RuleElement, 1);
 	    	    			newelement2->type = TERM2FREE;
                            newelement2->string = g_new (char, size + 1);
                            newelement2->string[size] = '\0';
                            while (size--)  newelement2->string[size] = c[size];
                            c += ((SplitElement *)split->data)->size;
     					    (g_list_nth (ret->binary, ((SplitElement *)split->data)->twin))->data = (gpointer) newelement2;
                            split=split->next;
                            }
				    }
                else  flag = 1; // if no reply then flag for quit
			    }


            else if (TERM == ((RuleElement *) ret->ruletext->data)->type) {              // its a term
				c = (char *) ((RuleElement *) ret->ruletext->data)->string;
                while (' '  == *c) c++;
                if ('\t' != *c)                 // non tabbed string
                  while (!flag){                // untill error
                    while (' '  == *c) c++;     // skip spaces
                    if (!*c) break;             // end of rule!
                    while (' '  == *ret->text) ret->text++;
                    if (*(ret->text) != *c){    // the letters are not thesame!
                        if ( (*c        > 0x40) &&      // second chance
                             (*ret->text > 0x40) &&
                             ((*c        & 0x1f) < 0x1b) &&
                             ((*ret->text & 0x1f) < 0x1b) &&
                             (*c & 0x1f) == (*ret->text & 0x1f)  ){     // case insensitive compare
                                c++;
                                ret->text++;    // advance
                                }
                        else flag=1;            // else error
                        }
                    else { c++; ret->text++; }  // advance
                    }
                else {
                    if ('f'==*(c+1))        // tabbing string with force space check
                        if (' ' != *ret->text)  // if no space then error
                           flag=1;
                    }
                }
                
                
                
                                            // finished all rules
			if (flag) {                     // errors occurred
                if (todolist) {             // stuff to do?
                    g_free(ret);
                    ret = todolist->data;   // get next state
                    todolist=g_list_remove(todolist, todolist->data);   // remove from list
                    }
                else break;
                }
			ret->ruletext = ret->ruletext ->next;   // next rule piece
                
            }
          }
        rules=rules->next;                          // next rule
	    }
    g_free(ret);
	return retlist;
}




/*
 * Asm asm_assembleold (char* text, GList* rules, int address, int last, int count)
 * {
 *  Asm error;
 *  Asm ret;
 *  Asm rep;
 *  GList *ruletext;
 *  RuleElement *newelement;
 *  int flag;
 *  int size;
 *  char* start;
 *  char* c;
 *  GList* tried = NULL;
 *  GList* current = NULL;    
 *     
 *  error.text = (char*) &dis_error;
 *  error.binary = &dis_error;
 * 
 * while (rules){
 * 	    ret.text = text;
 *         flag=1;
 * 	    if (TRANRULE != ((Rule *) rules->data)->type){
 *             while (' '  == *ret.text) ret.text++;
 *             if (!isxdigit(*ret.text))    break;
 *             if (count)    break;
 *             flag=0;
 *             while (isxdigit(*ret.text)){
 *                    if (*ret.text< 0x3A)     flag = (flag<<4) + *ret.text - 0x30;
 *                    else                     flag = (flag<<4) + (*ret.text & 7) + 9;
 *                    ret.text++;
 *                    }
 * 			if (    (RELATIVERULE == ((Rule *) rules->data)->type)
 *                 || (URELATIVERULE == ((Rule *) rules->data)->type))
 * 				    flag -= address;
 *             {
 *                 GList *number = ((Rule *) rules->data)->bit->next;
 *                 GList *opp = ((Rule *) rules->data)->asci;
 *                 int imm;
 *                     while(opp){
 *                         imm=GPOINTER_TO_INT (number->data);
 *                         c= (char*)opp->data;
 *                         if ('~' == *c) {
 *                             imm = flag;
 *                             flag = (int) number->data;
 *                             c++;
 *                             }
 *                         switch (*c){
 *                             case '+':
 *                                 flag +=imm;
 *                                 break;
 *                     
 *                             case '-':
 *                                 flag -=imm;
 *                                 break;
 *             
 *                             case '*':
 *                                 flag *=imm;
 *                                 break;
 *                     
 *                             case '/':
 *                                 flag /=imm;
 *                                 break;
 *             
 *                             case '>':
 *                                 flag = flag >> imm;
 *                                 break;
 *                                 
 *                             case '<':
 *                                 flag = flag << imm;
 *                                 break;
 *                                 
 *                             case '@':
 *                                 flag = ((flag << imm) | (flag >> (int)((Rule *) rules->data)->bit->data - imm)) & (1 << (int)((Rule *) rules->data)->bit->data);
 *                                 break;
 *                                 
 *                             default:
 *                                 break;
 *                             }
 *                         opp=opp->next;
 *                         number=number->next;
 *                         }
 *             }
 *             
 *             start = g_new(char,(int) ((Rule *) rules->data)->bit->data);
 *             c= start + (int) ((Rule *) rules->data)->bit->data;
 *             *c='\0';
 *             while(start != c--){
 *                 if (flag & 1)   *c = 'I';
 *                 else            *c = 'O';
 *                 flag = flag >> 1;
 *                 }
 * 			newelement = g_new (RuleElement, 1);
 * 			newelement->string = start;
 * 			newelement->type = TERM2FREE;
 * 			ret.binary = g_list_append (NULL, newelement);
 *             
 *             while (' '  == *ret.text) ret.text++;
 *             if (last && *ret.text)  break;
 *             else                    return ret;
 * 		    }
 *         else {
 * 	      ret.text = text;
 *           current = tried;
 *           ruletext=((Rule *) rules->data)->asci;
 * 		  ret.binary = g_list_copy (((Rule *) rules->data)->bit);
 *         
 * 		  while (1){
 * 		    flag = 0;
 *             while (' '  == *ret.text) ret.text++;
 * 		    if (!ruletext) 
 *                 if (last && *ret.text)  break;
 *                 else        
 *                     if (!count--)   return ret;
 *                     else    {
 *  	                    ret.text = text;
 *                         current = tried;
 *                         ruletext=((Rule *) rules->data)->asci;
 * 		                ret.binary = g_list_copy (((Rule *) rules->data)->bit);
 *                         break;
 *                         }
 *             
 *             
 * 		    if (NONTERM == ((RuleElement *) ruletext->data)->type){
 *                 if (!current){
 *                     tried = g_list_append (tried,(int) 0);
 *                     current = g_list_last(tried);
 * 				    rep = asm_assembleold (ret.text, ((RuleElement *) ruletext->data)->rules, address, last && !ruletext->next,(int) current->data);
 *                     }
 *                 else {
 *                     if (!current->next) ((int)current->data)++;
 *                     else if (-1 == (int)current->next->data) {((int)current->data)++;(int)current->next->data=0;}
 *                     rep = asm_assembleold (ret.text, ((RuleElement *) ruletext->data)->rules, address, last && !ruletext->next,(int) current->data);
 *                     }
 *                 
 *                     
 *  				if (&dis_error != rep.binary){
 *                     ret.text = rep.text;
 * 					newelement = g_new (RuleElement, 1);
 * 					newelement->rules = rep.binary;
 * 					newelement->type = NONTERM2FREE;
 * 					(g_list_nth (ret.binary, ((RuleElement *) ruletext->data)->twin))->data = (gpointer) newelement;
 * 				    }
 *                 else {  
 *                     flag = 1; 
 *                     ((int)current->data) = -1;
 *                     }
 *                current=current ->next;
 * 			    }
 *             else {
 * 				c = (char *) ((RuleElement *) ruletext->data)->string;
 *                 if ('\t' != *c) while (!flag){
 *                     while (' '  == *c) c++;
 *                     while (' '  == *ret.text) ret.text++;
 *                     if (!*c) break;
 *                     if (*(ret.text) != *c){
 *                         if ( (*c        > 0x40) &&
 *                              (*ret.text > 0x40) &&
 *                              ((*c        & 0x1f) < 0x1b) &&
 *                              ((*ret.text & 0x1f) < 0x1b) &&
 *                              (*c & 0x1f) == (*ret.text & 0x1f)  ){
 *                                 c++;
 *                                 ret.text++;
 *                                 }
 *                         else flag=1;
 *                         }
 *                     else { c++; ret.text++; }
 *                     }
 *                     
 *                 }
 * 			if (flag) break;
 * 			ruletext = ruletext ->next;
 *             }
 * 		  asm_freestringlist (ret.binary);
 *           }
 * 	    if (NULL == tried  || -1 == (int) tried->data){
 *                     rules = rules->next;
 *                     tried = NULL; 
 *                     current = NULL;
 *                     }
 * 
 * 	    }
 * 	return error;
 * }
 */


GList* asm_copystringlist (GList* list)         // copy an asm list
{
 GList *scan, *split;
 SplitElement *splitelement;
 RuleElement* newrule;
 list = g_list_copy(list);                      // copy the list
 scan = list;
 while (scan && &dis_error != scan)             // walk list
	{
		if (NONTERM2FREE == ((RuleElement *) scan->data)->type) // if the nonterm needs to be freed then make an exact copy
		{
            newrule = g_new(RuleElement, 1);                    // make a new nonterm2free list
            newrule->type = NONTERM2FREE;
            newrule->string = ((RuleElement *) scan->data)->string;    
            newrule->rules = asm_copystringlist (((RuleElement *) scan->data)->rules);  // just copy the old one
            scan->data = newrule;
		}
        
		if (TERM2FREE == ((RuleElement *) scan->data)->type)    // if its a term that needs to be freed then 
		{
            newrule = g_new(RuleElement, 1);                    // make a copy
            newrule->type = TERM2FREE;
            newrule->string =  g_strdup(((RuleElement *) scan->data)->string); // duplicate the string
            newrule->rules = NULL;
            scan->data = newrule;
        }

		if (SPLITNONTERM2FREE == ((RuleElement *) scan->data)->type)    // if its a split non term that needs to be freed then 
		{
            newrule = g_new(RuleElement, 1);                    // make a new splitnonterm2free list
            newrule->type = NONTERM2FREE;
            newrule->string = ((RuleElement *) scan->data)->string;    
            newrule->rules = asm_copystringlist (((RuleElement *) scan->data)->rules);  // just copy the old one
            newrule->split = g_list_copy(((RuleElement *) scan->data)->split);
            split=newrule->split;
            while (split){
                if (SPLIT2FREE == ((SplitElement *)split->data)->type){
                    splitelement = g_new(SplitElement,1);
                    splitelement->string=g_strdup(((SplitElement *)split->data)->string);
                    splitelement->type=SPLIT2FREE;
                    }
                }
            scan->data = newrule;
        }
        
		scan = scan->next;      // next
	}
    return list;                // return the copy
}

void asm_freestringlist (GList* list)   // free a asm list
{
	GList *scan = list;
    GList * split;

	while (scan && &dis_error != scan)      // walk the list
	{
		if  (SPLITNONTERM2FREE == ((RuleElement *) scan->data)->type)  {   // on splitnonterm2frees
			asm_freestringlist (((RuleElement *) scan->data)->rules);   // free their rules
            split = ((RuleElement *) scan->data)->split;
            while (split) {
                if (SPLIT2FREE == ((SplitElement *)split->data)->type) {    // free all split2frees
//                	g_list_free (((SplitElement *)split->data)->bin);         // free the list
                    g_free (split->data);
                    }
                split=split->next;
                }
			g_free (scan->data);                                        // and the element
		    }
            
		else if (NONTERM2FREE == ((RuleElement *) scan->data)->type) {     // on nonterm2frees
			asm_freestringlist (((RuleElement *) scan->data)->rules);   // free their contents
			g_free (scan->data);                                        // and element (remember not the string!!!)
    		}
            
		else if (TERM2FREE == ((RuleElement *) scan->data)->type) {  // on term2free
			g_free (((RuleElement *) scan->data)->string);          // free the string
			g_free (scan->data);                                    // and element
		    }
        
		scan = scan->next;      // next
	}

	g_list_free (list);         // free the list
}


GList *uchararr2glist (int size, uchar * vals)      // turn a uchararr into a glist of 0 or 1 gpointers
//                  number of chars, the chararr
{
	GList *ret = NULL;
	uchar c;
	int count;

	while (size--)
	{
		c = *vals;
		vals++;
		for (count = 8; count; count--)     // 8 bits per char
		{			ret = g_list_append (ret, GINT_TO_POINTER (c & 1));
			c >>= 1;                        // keep sticking them on
		}
	}
	return ret;
}

void asm_print (GList* list)                // print the asmlist
{
	if (&dis_error == list)                 // if error then undefined
		g_print ("Undefined");
	else
		while (list)                        // walk list
		{
			if (TERM == ((RuleElement *) list->data)->type ||       // print terms
                TERM2FREE == ((RuleElement *) list->data)->type)
				g_print ("%s", ((RuleElement *) list->data)->string);
			else
				asm_print (((RuleElement *) list->data)->rules);    // recurse into nonterms
			list = list->next;
		}

}

char* asm_sprint (GList* list, char* replystring)
            // asm list, mallocked string
{
 char* oldreplystring;
		while (list)    // walk list
		{
		if (TERM == ((RuleElement *) list->data)->type ||       // print all terms
            TERM2FREE == ((RuleElement *) list->data)->type) {
                oldreplystring = replystring;
                replystring = asm_sprint_tabstring(((RuleElement *)list->data)->string, replystring);
                                            // call sprint_tabstring
                g_free(oldreplystring);     // free the old string
                }
		else    replystring = asm_sprint (((RuleElement *) list->data)->rules, replystring); //else recurse
		list = list->next;                  // next
		}
/*                  Replaced by usng the tabstring function
 *  oldreplystring = replystring;
 *  while (oldreplystring)
 *         if ('\\' == *oldreplystring && 't' == *(oldreplystring+1)) {
 *                 char* temp = replystring;
 *                 value=0;
 *                 *oldreplystring='\0';
 *                 oldreplystring += 2;
 *                 while ( '0' >= *oldreplystring &&
 *                         '9' <= *oldreplystring ){
 *                             value = value * 10 + (*oldreplystring - '0');
 *                             oldreplystring++;
 *                             }
 *                 if (*oldreplystring) oldreplystring++;
 *                 g_strlen(replystring);
 *                 }
 */
                
                

 return replystring;
}

char* asm_sprint_tabstring (char* tabstring, char* replystring)
//                          string to append, original string
{
 int value;
 int value2=0;
 char* temp;
 if ('\t' != *tabstring)  return g_strconcat(replystring, tabstring, NULL);     // if its a tab string
 if ('f' == *(tabstring+1)) {if (*(tabstring+2)) value2 = g_strtod (tabstring+2, &temp);}
 else                       {if (*(tabstring+2)) value2 = g_strtod (tabstring+1, &temp);}
            // get the tab offset into value2
            
 value = strlen(replystring);
 if (value2) value = value2 - (value % value2); // tab to offset
 else value = 0;
 
 if (value<=0) return g_strconcat(replystring, tabstring, NULL);    // if no tab then just concat
 temp = g_new(char, value + 1);
 temp[value] = '\0';
 while (value) temp[--value] = ' ';                                 // else stick spaces on
 replystring = g_strconcat(replystring, temp, NULL);
 g_free (temp);
 return replystring;
}


int asm_sbprint (GList* list, uchar** data)        // get binaryasm list into a chararr
                // asmlist, chararr ptr ptr
{
    char* text = asm_sprint (list, g_new0(char,1));     // print list into a string
    int size = strlen(text);                            // get length
    char* c = text + size;                              // work from the end
    if (size%8) size = (size>>3) + 1;                   // if overflow then round up
    else        size =  size>>3;
    *data = g_new0(char, size);                         // allocate with 0s
    size = 0;
    while (c-->text){                                   // walk string and stick 1s at points
        if ('X'==*c  || 'I'==*c) ((*data)[size>>3]) |= 1 << (size&7);
        size++;
        }
    return (size>>3);                                   // return rounded down
}

char* asm_symboltable_find_address (Symbol* symbol_table, int symbol_count,long address)
{
 int index;
 for (index=0; index<symbol_count; index++){
    if (symbol_table[index].value == address)
        return symbol_table[index].name;
    }
 return NULL;
}


int asm_getsymbol_length (char* c)
{
 int ret=0;
 while (('_' == *c) ||
        ('a' <= *c && 'z' >= *c) ||
        ('Z' <= *c && 'Z' >= *c) ||
        ('1' <= *c && '9' >= *c && ret)) {ret++; c++;}
 return ret;
}


int asm_findstrinsymboltable(char* string, Symbol* array,int arrcount)
{
  char* mystr = g_strdup(string);  /*find string in an array of strings and return position*/
  int temp1=0 ,temp2=0;
  while (mystr[temp1]) if (' ' == mystr[temp1]) temp1++;
                       else mystr[temp2++] = mystr[temp1++];
  mystr[temp2]='\0';
  while (arrcount--){
	if (!g_strcasecmp(mystr,array[arrcount].name)) break;
    }
  g_free(mystr);
  return arrcount;
}
